﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using TwinCAT.Ads;

namespace PlcDataManager
{
    public enum PLCDataOperationCode
    {
        ADS_FAILED = -1,
        ADS_SUCCESS,
        ADS_DATAFORMAT
    }

    public class PlcDataDriverClass
    {
        private TcAdsClient adsClient;
        private string plcLocalName = null;
        private string plcLocalAmsNetId = null;
        private int plcPort = 851;

        private bool plcConectioed = false;
        private ArrayList notificationHandles;
        private Dictionary<string, int> variableHandles = new Dictionary<string, int>();
        private Dictionary<string, TcAdsSymbolInfo> adsVariableInfos = new Dictionary<string, TcAdsSymbolInfo>();
        private List<TcAdsSymbolInfo> adsSymbolInfos;
        private object threadLockObject = new object();

        public delegate void ConnectionEventHandler(Boolean isConnected);
        public event ConnectionEventHandler ConnectionEstablished;

        public delegate void ChangeNotificationEventHandler(string tag, object newValue);
        public event ChangeNotificationEventHandler ChangeUpdate;

        public Boolean IsPLCConnected
        {
            get
            {
                lock (threadLockObject)
                {
                    return plcConectioed;
                }
            }
            set
            {
                lock (threadLockObject)
                {
                    plcConectioed = value;
                }
            }
        }

        private Type GetTypeForTag(string tag)
        {
            TcAdsSymbolInfo info = null;
            if (adsVariableInfos.TryGetValue(tag.ToUpper(), out info))
                return GetSystemType(info.Datatype);
            else
                throw new ApplicationException("Tag not known " + tag);
        }

        private int GetHandle(string tag)
        {
            int handle = 0;
            if (!variableHandles.ContainsKey(tag))
            {
                handle = 0;
            }
            else
            {
                handle = variableHandles[tag];
            }
            return handle;
        }

        public PLCDataOperationCode WriteValue(string tag, object value)
        {
            PLCDataOperationCode ret = PLCDataOperationCode.ADS_FAILED;
            object rawObjectValue = value;
            Type valueType = rawObjectValue.GetType();
            Type tagType = GetTypeForTag(tag);
            int handle = GetHandle(tag);
            if (tagType == typeof(System.String))
            {
                string myObjectString = value.ToString();
                adsClient.WriteAny(handle, rawObjectValue, new int[] { myObjectString.Length });
                ret = PLCDataOperationCode.ADS_SUCCESS;
            }
            else
            {
                adsClient.WriteAny(handle, rawObjectValue);
                ret = PLCDataOperationCode.ADS_SUCCESS;
            }
            return ret;
        }

        public PLCDataOperationCode ReadValue(string tag, ref object value)
        {
            PLCDataOperationCode ret = PLCDataOperationCode.ADS_FAILED;
            if (IsPLCConnected)
            {
                int handle = GetHandle(tag);
                Type type = GetTypeForTag(tag);
                try
                {
                    if (type == typeof(System.String))
                    {
                        value = adsClient.ReadAny(handle, typeof(String), new int[] { 80 }).ToString();
                    }
                    else
                    {
                        value = adsClient.ReadAny(handle, type);
                    }
                    ret = PLCDataOperationCode.ADS_SUCCESS;
                }
                catch (Exception ex)
                {
                    ret = PLCDataOperationCode.ADS_FAILED;
                }
            }
            return ret;
        }

        public PLCDataOperationCode InitializeConnectionToPLC(string plcName, string amsNetID, int port)
        {
            PLCDataOperationCode ret = PLCDataOperationCode.ADS_FAILED;
            plcLocalName = plcName;
            plcLocalAmsNetId = amsNetID;
            plcPort = port;

            if (adsClient == null)
            {
                adsClient = new TcAdsClient();
                notificationHandles = new ArrayList();
                adsClient.AdsNotificationEx += new AdsNotificationExEventHandler(adsClient_AdsNotificationEx);
            }
            try
            {
                adsClient.Connect(plcLocalAmsNetId, plcPort);
            }
            catch (Exception ex)
            {
                ret = PLCDataOperationCode.ADS_FAILED;
            }

            if (adsClient.IsConnected == true)
            {
                try
                {
                    LoadAllAdsSymbolInfos();
                }
                catch (Exception ex)
                {
                    IsPLCConnected = false;
                    throw new ApplicationException("TwinCAT run mode? " + ex.Message);
                }


                for (int i = 0; i < adsSymbolInfos.Count; i++)
                {
                    TcAdsSymbolInfo currentSymbol = adsSymbolInfos[i];
                    int handle = adsClient.CreateVariableHandle(currentSymbol.Name);
                    variableHandles.Add(currentSymbol.Name, handle);
                    Type t = GetType(currentSymbol.Type);
                    if (t == null)
                    {
                        continue;
                    }
                    if (t == typeof(String))
                    {
                        notificationHandles.Add(adsClient.AddDeviceNotificationEx(currentSymbol.Name, AdsTransMode.OnChange, 100, 0, handle, t, new int[] { 80 }));
                    }
                    else
                    {
                        notificationHandles.Add(adsClient.AddDeviceNotificationEx(currentSymbol.Name, AdsTransMode.OnChange, 100, 0, handle, t));
                    }
                }
                IsPLCConnected = true;
                if (ConnectionEstablished != null)
                {
                    ConnectionEstablished(true);
                }
                ret = PLCDataOperationCode.ADS_SUCCESS;
            }
            return ret;
        }


        public void CloseAdsConnection()
        {
            adsClient.AdsNotificationEx -= adsClient_AdsNotificationEx;
            adsClient.Dispose();
            notificationHandles.Clear();
            Thread.Sleep(2000);
            IsPLCConnected = false;
        }


        private static Type GetSystemType(AdsDatatypeId adsDatatypeId)
        {
            switch (adsDatatypeId)
            {
                case AdsDatatypeId.ADST_UINT8: return typeof(System.Byte);
                case AdsDatatypeId.ADST_BIT: return typeof(System.Boolean);
                case AdsDatatypeId.ADST_INT16: return typeof(System.Int16);
                case AdsDatatypeId.ADST_INT32: return typeof(System.Int32);
                case AdsDatatypeId.ADST_REAL32: return typeof(System.Single);
                case AdsDatatypeId.ADST_REAL64: return typeof(System.Double);
                case AdsDatatypeId.ADST_INT8: return typeof(System.SByte);
                case AdsDatatypeId.ADST_STRING: return typeof(System.String);
                default:
                    throw new ApplicationException("Unsupported type?");
            }
        }

        private Type GetType(string aType)
        {
            switch (aType)
            {
                case "BOOL":
                    return typeof(Boolean);

                case "INT":
                    return typeof(Int16);

                case "REAL":
                    return typeof(float);

                case "STRING(80)":
                    return typeof(String);
                default:
                    throw new ApplicationException("Unsupported type?");
            }
        }

        private void adsClient_AdsNotificationEx(object sender, AdsNotificationExEventArgs e)
        {
            string tag = null;
            int handle = (int)e.UserData;
            tag = variableHandles.FirstOrDefault(x => x.Value == handle).Key;
            if (tag != null && ChangeUpdate != null)
            {
                ChangeUpdate(tag, e.Value);
            }
        }


        private List<TcAdsSymbolInfo> LoadAllAdsSymbolInfos()
        {
            if (adsSymbolInfos == null)
            {
                this.adsSymbolInfos = new List<TcAdsSymbolInfo>();
                TcAdsSymbolInfoLoader loader = adsClient.CreateSymbolInfoLoader();
                LoadAllSubSymbols(loader.GetSymbols(true));
            }
            return adsSymbolInfos;
        }

        private void LoadAllSubSymbols(TcAdsSymbolInfoCollection adsSymbolInfoCollection)
        {
            foreach (TcAdsSymbolInfo symbol in adsSymbolInfoCollection)
            {
                if (symbol.Name.Contains(plcLocalName) == true && symbol.Name.Contains("myTimer") == false)
                {
                    adsVariableInfos.Add(symbol.Name.ToUpper(), symbol);
                    adsSymbolInfos.Add(symbol);
                    LoadAllSubSymbols(symbol.SubSymbols);
                }
            }
        }
    }
}
